// Copyright © 2011 Kris Maglione <maglione.k@gmail.com>
//
// This work is licensed for reuse under an MIT license. Details are
// given in the LICENSE file included with this file.
"use strict";

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cu = Components.utils;

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

try {
    var { Services, _, log, prefs } = Cu.import("resource://exttest/exttest.jsm", {});
}
catch (e) {
    var { Services, _, log, prefs } = Cu.import(Components.stack.filename.replace(/[^\/]*$/, "") + "../modules/exttest.jsm", {});
}

function EvalCheck(name, test) {
    this.name = name;
    this.id = EvalCheck.id++;
    if (test)
        this.test = test;

    function Check() {
        this.locations = {};
    }
    Check.prototype = this;
    return Check;
}
EvalCheck.id = 0;
EvalCheck.prototype = {
    get contractID()        "@dactyl.googlecode.com/exttest/eval-check/" + this.name,
    get classID()           Components.ID("{27521dbf-8182-4ce7-b0f6-2b7d3a25d4c" + this.id + "}"),
    get classDescription()  "Extension Test " + this.name + " check injector",
    QueryInterface:         XPCOMUtils.generateQI([Ci.nsIDOMGlobalPropertyInitializer]),

    test: function test(arg) {
        if (typeof arg !== "function")
            return "setTimeout.eval";
        return false
    },

    check: function check(func, args) {
        try {
            let res = this.test.apply(this, args);
            if (res) {
                let caller = Components.stack.caller.caller;
                let loc = caller.filename + ":" + caller.lineNumber;

                if (!this.locations.hasOwnProperty(loc) && prefs.console.get(this.name, true))
                    log(_("unsafeUse", this.name) +
                        (typeof res !== "boolean" ? " (" + _.apply(null, Array.concat(res)) + ")" : ""),
                        null, caller);

                this.locations[loc] = 1 + (this.locations[loc] || 0);
                func.exttestLocations[loc] = 1 + (func.exttestLocations[loc] || 0);
            }
        }
        catch (e) {
            Cu.reportError(e);
        }
    },

    init: function init(window) {
        let self = this;

        result.exttestLocations = {};
        function result() {
            self.check(result, arguments);
            return Components.lookupMethod(window, self.name).apply(this, arguments);
        }
        return result;
    }
};

function XMLHttpRequest() {
    let xhr = Components.classesByID["{d164e770-4157-11d4-9a42-000064657374}"]
                        .createInstance(Ci.nsIXMLHttpRequest);

    let self = this;
    let origOpen = xhr.open;
    xhr.open = function open() {
        if (Services)
            self.check(Services.Xmlhttp, arguments);
        return origOpen.apply(this, arguments);
    };
    return xhr;
}
if (Services)
    Services.Xmlhttp.exttestLocations = {};
XMLHttpRequest.prototype = {
    __proto__:  EvalCheck.prototype,
    classID:    Components.ID("{770002b4-5c23-48f6-9665-454978bb3774}"),
    contractID: "@mozilla.org/xmlextras/xmlhttprequest;1",
    name:       "XMLHttpRequest",

    locations: {},

    test: function test(method, url, async, user, pass) {
        let uri = Services.io.newURI(url, null, null);
        while (uri instanceof Ci.nsIJARURI)
            uri = uri.JARFile;

        let proto = Services.io.getProtocolHandler(uri.scheme);
        if (!async && !(proto.protocolFlags & Ci.nsIProtocolHandler.URI_IS_LOCAL_RESOURCE))
            return ["xmlhttp.sync", url];
        if (pass && uri.scheme !== "https")
            return "xmlhttp.pass";
        if (uri instanceof Ci.nsIURL && /(^|&)pass/.test(uri.QueryInterface(Ci.nsIURL).query))
            return "xmlhttp.passQuery";
        return false;
    },
};

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory([
        new EvalCheck("setInterval"),
        new EvalCheck("setTimeout"),
        XMLHttpRequest
    ]);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule([XMLHttpRequest]);

