// Copyright © 2010-2011 Kris Maglione <maglione.k@gmail.com>
//
// This work is licensed for reuse under an MIT license. Details are
// given in the LICENSE file included with this file.
"use strict";

if (true) {
    let {
        ADDON_ID, Cc, Ci, Cu, Services, XUL, _, categories,
        iterCategory, getOwnPropertyNames, log, module, prefs,
        wrap, xmlToDom, xpath
    } = Components.utils.import("resource://exttest/exttest.jsm", {});

    let file = document.documentURI.replace(/.*\//, "");

    let $ = function $(selector, node) (node || document).querySelector(selector);

    // Clear legacy preferences
    ["IDs", "properties"].forEach(function (pref) { prefs.reset(pref); });

    if (prefs.safe.has("Array"))
        ["Array", "Boolean", "Date", "Function", "Number", "Object", "RegExp", "String"]
            .forEach(function (pref) {
                prefs.safe.set(pref + ".prototype", prefs.safe.get(pref));
                prefs.safe.reset(pref);
            });

    // Touch as many known properties as possible, since they 
    // tend not to show up until we do, at which point they appear
    // as loose properties

    // Only a problem on Gecko 1.x
    if (!Object.getOwnPropertyNames)
        for (let k in Cc)
            Cc[k];

    for each (let iface in Ci)
        if (iface instanceof Ci.nsIJSIID && window instanceof iface)
            window.QueryInterface(iface);

    for (let k in Ci)
        if (/^nsIDOM/.test(k))
            if (k.substr(6) in window)
                window[k.substr(6)];
    for (let k in Ci)
        if (/^nsI/.test(k))
            if (k.substr(3) in window)
                window[k.substr(3)];

    for (let [, cat] in Iterator(categories))
        for (let property in iterCategory(cat))
            if (property in window)
                window[property];

    ["Application", "Array", "ArrayBuffer", "AttributeName", "Boolean",
     "Components", "CSSFontFaceStyleDecl", "CSSGroupRuleRuleList",
     "CSSNameSpaceRule", "CSSRGBColor", "CSSRect", "ComputedCSSStyleDeclaration",
     "Date", "Error", "EvalError", "Float32Array", "Float64Array", "Function",
     "HTMLDelElement", "HTMLInsElement", "HTMLSpanElement", "Infinity",
     "InnerModalContentWindow", "InnerWindow", "Int16Array", "Int32Array",
     "Int8Array", "InternalError", "Iterator", "JSON", "KeyboardEvent", "Math",
     "NaN", "Namespace", "Number", "Object", "Proxy", "QName",
     "ROCSSPrimitiveValue", "RangeError", "ReferenceError", "RegExp",
     "StopIteration", "String", "SyntaxError", "TypeError", "URIError",
     "Uint16Array", "Uint32Array", "Uint8Array", "XML", "XMLHttpProgressEvent",
     "XMLList", "XMLSerializer", "XMLStylesheetProcessingInstruction", 
     "XPCNativeWrapper", "XPCSafeJSWrapper", "XULControllers", "_content",
     "constructor", "decodeURI", "decodeURIComponent", "encodeURI",
     "encodeURIComponent", "escape", "eval", "isFinite", "isNaN", "isXMLName",
     "parseFloat", "parseInt", "undefined", "unescape", "uneval"
    ].concat(getOwnPropertyNames({}.__proto__))
     .forEach(function (k) { if (k in window) window[k]; });

    let prevValues = {};
    let labels = {};

    let compare = function compare(name, shortName, values, global) {
        XML.ignoreWhitespace = true;

        let _seen = {};
        values = values.filter(function (val) !({}).hasOwnProperty.call(_seen, val) && (_seen[val] = true));

        let safeValues = values.sort();
        if (prefs.safe.has(name))
            safeValues = JSON.parse(prefs.safe.get(name));
        else
            prefs.safe.set(name, JSON.stringify(values));

        values = values.filter(function (p) safeValues.indexOf(p) === -1).sort();
        let newValues = values;
        if (name in prevValues)
            newValues = values.filter(function (v) prevValues[name].indexOf(v) == -1);
        prevValues[name] = values;

        if (button) {
            if (values.length && prefs.status.get(name, true)) {
                if (!(name in button.labels))
                    button.appendChild(xmlToDom(<toolbarbutton class="exttest-label toolbarbutton-1"
                                                               context="exttest-context-count"
                                                               oncommand="this.copy()"
                                                               key={name}
                                                               xmlns={XUL}>
                        <label class="exttest-label-title" value={(shortName || name[0]) + ":"}/>
                        <label class="exttest-label-count" key={name + "-count"}/>
                    </toolbarbutton>, document, button.labels));

                button.labels[name + "-count"].value = values.length;
                button.labels[name].hidden = false;
                button.labels[name].setAttribute("tooltiptext", wrap(values, _("newValues", name.replace(/\..*/, "")) + " ", " "));
                button.labels[name].setAttribute("value", values.join(" "));
                button.labels[name].copy = function copy() {
                    let data = this.getAttribute("value");
                    Services.clipboardHelper.copyString(data);
                };
                button.labels[name].reset = function copy() {
                    prefs.safe.reset(this.getAttribute("key"));
                    this.hidden = true;
                };
            }
            else if (name in button.labels) {
                button.labels[name].hidden = true;
            }
        }

        if (!global && newValues.length && prefs.console.get(name, true))
            log(_("addedTo", name), newValues);
    };

    let ignoreGlobals = {
        __SS_dyingCache: true,
        __SS_lastSessionWindowID: true,
        __SS_restoreID: true,
        __SS_tabsToRestore: true,
        __SSi: true
    };

    let button;
    let sandbox = Cu.Sandbox(window);
    let check = function check() {
        button = $("#addon-bar > #exttest-toolbar, #exttest-status");
        if (button && !button.labels)
            button.labels = {};

        // Global properties
        compare("global." + file, "g", [].concat.apply([], [window, window.__proto__, sandbox]
                                         .map(getOwnPropertyNames))
                                         .filter(function (p) !(/\s/.test(p) || ignoreGlobals.hasOwnProperty(p))));

        // DOM node IDs
        compare("IDs." + file, "ID", [a.value for (a in xpath("//@id", document))
                                      if (!/^spring\d+$/.test(a.value))]);

        // Categories
        let cats = Array.concat.apply(null,
            categories.map(function (cat) [cat.replace(/\s/g, "-") + ":" + prop
                                           for (prop in iterCategory(cat))]));
        compare("categories", "cat", cats);

        compare("Math", "M", getOwnPropertyNames(Math));

        // Object prototypes
        ["Array",
         "Boolean",
         "Date",
         "Function",
         "Number",
         "Object",
         "RegExp",
         "String",
         "XMLHttpRequest"].forEach(function (name) {
            compare(name + ".prototype", name[0] + ".p", getOwnPropertyNames(window[name].prototype));
        });

        // XPCOM classes and interfaces
        [["Cc", Cc],
         ["Ci", Ci]].forEach(function ([name, obj]) {
            compare(name, name, getOwnPropertyNames(obj));
        });

        ["setInterval", "setTimeout"].forEach(function (name) {
            if ("exttestLocations" in window[name]) {
                compare(name, name[0]+name[3],
                        getOwnPropertyNames(window[name].exttestLocations),
                        true);
            }
        });

        if ("exttestLocations" in Services.Xmlhttp)
            compare("XMLHttpRequest", "XHR",
                    getOwnPropertyNames(Services.Xmlhttp.exttestLocations),
                    true);
    };

    ["setInterval", "setTimeout", "XMLHttpRequest"].forEach(function (name) {
        if (!prefs.safe.has(name))
            prefs.safe.set(name, "[]");
    });

    addEventListener("load", function onLoad() {
        removeEventListener("load", onLoad, false);

        function firstRun() {
            let addonBar = $("#addon-bar");
            if (addonBar && !$("#exttest-toolbar")) {
                addonBar.insertItem("exttest-toolbar");
                document.persist(addonBar.id, "currentset");
            }
        }

        function checkVersion(addon) {
            if (prefs.get("version", null) != addon.version)
                firstRun();

            prefs.set("version", addon.version);
        }

        if ("nsIExtensionManager" in Ci)
            checkVersion(Services.extensionManager.getItemForID(ADDON_ID))
        else {
            // Import it into the global scope to prevent loose variable
            // warnings when other extensions import it.
            Components.utils.import("resource://gre/modules/AddonManager.jsm");
            AddonManager.getAddonByID(ADDON_ID, checkVersion);
        }

        Array.forEach($("#exttest-context-button").children, function (child) {
            if (child.hasAttribute("test"))
                child.hidden = !(child.getAttribute("test") in window);
        });

        let copy = $("#exttest-context-copy");
        if ($("#context-copy"))
            copy.style.listStyleImage = window.getComputedStyle($("#context-copy"), null).listStyleImage;

        (function timeout() {
            try {
                check();
            }
            catch (e) {
                Cu.reportError(e);
            }
            setTimeout(timeout, prefs.get("interval"));
        })();
    }, false);

    /*
        XMLHttpRequest().open("GET", "chrome://exttest/content/exttest.js", false);
        XMLHttpRequest().open("GET", "http://www.google.com", false);
        XMLHttpRequest().open("GET", "http://www.google.com", true, "foo", "bar");
        XMLHttpRequest().open("GET", "https://www.google.com", true, "foo", "bar");
        XMLHttpRequest().open("GET", "http://www.google.com", true);
        XMLHttpRequest().open("GET", "http://www.google.com/?pass=foo", true);
        XMLHttpRequest().open("GET", "http://www.google.com/?user=foo&pass=bar", true);
        clearInterval(setInterval("12", 12));
        setTimeout("12", 12);
    */
}
