// Copyright © 2009-2011 Kris Maglione <maglione.k@gmail.com>
//
// This work is licensed for reuse under an MIT license. Details are
// given in the LICENSE file included with this file.
"use strict";

let EXPORTED_SYMBOLS = [];

let ADDON_ID = "extension-test@dactyl.googlecode.com";

let Cc = Components.classes;
let Ci = Components.interfaces;
let Cu = Components.utils;
let XUL = Namespace("xul", "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul");

function _(key) {
    try {
        return strings.formatStringFromName(key, Array.slice(arguments, 1), arguments.length - 1);
    }
    catch (e) {
        return null;
    }
}

function module(uri) {
    let object = {};
    Cu.import(uri, object);
    return object;
}

var Services = {
    init: function () {
        this.add("catman",              "@mozilla.org/categorymanager;1",           Ci.nsICategoryManager);
        this.add("clipboardHelper",     "@mozilla.org/widget/clipboardhelper;1",    Ci.nsIClipboardHelper);
        this.add("console",             "@mozilla.org/consoleservice;1",            Ci.nsIConsoleService);
        this.add("debugger",            "@mozilla.org/js/jsd/debugger-service;1",   Ci.jsdIDebuggerService);
        this.add("extensionManager",    "@mozilla.org/extensions/manager;1",        Ci.nsIExtensionManager);
        this.add("io",                  "@mozilla.org/network/io-service;1",        Ci.nsIIOService);
        this.add("pref",                "@mozilla.org/preferences-service;1",       [Ci.nsIPrefBranch2, Ci.nsIPrefService]);
        this.add("runtime",             "@mozilla.org/xre/runtime;1",               [Ci.nsIXULAppInfo, Ci.nsIXULRuntime]);
        this.add("stringBundle",        "@mozilla.org/intl/stringbundle;1",         Ci.nsIStringBundleService);
        this.add("versionCompare",      "@mozilla.org/xpcom/version-comparator;1",  Ci.nsIVersionComparator);

        this.addClass("ScriptError",    "@mozilla.org/scripterror;1",               Ci.nsIScriptError, "init");
        this.addClass("String",         "@mozilla.org/supports-string;1",           Ci.nsISupportsString, "data");
        this.addClass("Xmlhttp",        "@mozilla.org/xmlextras/xmlhttprequest;1",  Ci.nsIXMLHttpRequest);
    },
    reinit: function () {},

    _create: function (classes, ifaces, meth, init, args) {
        try {
            var res = Cc[classes][meth || "getService"]();
            Array.concat(ifaces).forEach(function (iface) res.QueryInterface(iface));
        }
        catch (e) {
            return null;
        }

        if (init && args.length) {
            try {
                var isCallable = typeof res[init] === "function";
            }
            catch (e) {} // Ugh.

            if (isCallable)
                res[init].apply(res, args);
            else
                res[init] = args[0];
        }
        return res;
    },

    add: function add(name, class_, ifaces, meth) {
        const self = this;
        this.__defineGetter__(name, function () {
            delete self[name];
            return self[name] = self._create(class_, ifaces, meth);
        });
    },

    addClass: function (name, class_, ifaces, init) {
        const self = this;
        this[name] = function () self._create(class_, ifaces, "createInstance", init, arguments);
    },
};
Services.init();

let strings = Services.stringBundle.createBundle("chrome://exttest/locale/exttest.properties");

function xmlToDom(node, document, nodes) {
    if (node.length() != 1) {
        let domnode = document.createDocumentFragment();
        for each (let child in node)
            domnode.appendChild(xmlToDom(child, document, nodes));
        return domnode;
    }

    switch (node.nodeKind()) {
    case "text":
        return document.createTextNode(String(node));
    case "element":
        let domnode = document.createElementNS(node.namespace(), node.localName());
        for each (let attr in node.@*::*)
            domnode.setAttributeNS(attr.namespace(), attr.localName(), String(attr));

        for each (let child in node.*::*)
            domnode.appendChild(xmlToDom(child, document, nodes));
        if (nodes && node.@key)
            nodes[node.@key] = domnode;
        return domnode;
    default:
        return null;
    }
};

let WIDTH = 80;
let PREFIX = "    ";
function wrap(items, prefix, sep) {
    sep = sep || " ";
    let width = 0;
    let lines = [];
    lines.__defineGetter__("last", function () this[this.length - 1]);

    for (let [, item] in Iterator(items)) {
        if (item.length > width && (!lines.length || lines.last.length > 1)) {
            lines.push([prefix]);
            width = WIDTH - prefix.length;
            prefix = PREFIX;
        }
        width -= item.length + sep.length;
        lines.last.push(item, sep);
    }
    lines.last.pop();
    return lines.map(function (l) l.join("")).join("\n").replace(/\s+\n/gm, "\n");
}

function Prefs(branch, defaults) {
    this.constructor = Prefs; // Ends up Object otherwise... Why?

    this.branch = Services.pref[defaults ? "getDefaultBranch" : "getBranch"](branch || "");
    if (this.branch instanceof Ci.nsIPrefBranch2)
        this.branch.QueryInterface(Ci.nsIPrefBranch2);

    this.defaults = defaults ? this : new this.constructor(branch, true);
}
Prefs.prototype = {
    Branch: function Branch(branch) new this.constructor(this.root + branch),

    get root() this.branch.root,

    get: function get(name, defaultValue) {
        let type = this.branch.getPrefType(name);
        if (type === Ci.nsIPrefBranch.PREF_STRING)
            return this.branch.getComplexValue(name, Ci.nsISupportsString).data;
        if (type === Ci.nsIPrefBranch.PREF_INT)
            return this.branch.getIntPref(name);
        if (type === Ci.nsIPrefBranch.PREF_BOOL)
            return this.branch.getBoolPref(name);
        return defaultValue;
    },

    has: function has(name) this.branch.getPrefType(name) !== 0,

    getNames: function (branch) this.branch.getChildList(branch || "", { value: 0 }),

    set: function (name, value) {
        let type = typeof value;
        if (type === "string")
            this.branch.setComplexValue(name, Ci.nsISupportsString, Services.String(value));
        else if (type === "number")
            this.branch.setIntPref(name, value);
        else if (type === "boolean")
            this.branch.setBoolPref(name, value);
        else
            throw TypeError("Unknown preference type: " + type);
    },

    reset: function (name) {
        if (this.branch.prefHasUserValue(name))
            this.branch.clearUserPref(name);
    }
};

let prefs = new Prefs("extensions.exttest.");
prefs.console = prefs.Branch("console.");
prefs.safe = prefs.Branch("safe.");
prefs.status = prefs.Branch("status.");

// Categories which create globals
let categories = Array.concat.apply(null,
    ["global-constructor",
     "global-constructor-prototype-alias",
     "global-property",
     "global-privileged-property",
     "global-static-nameset",
     "global-dynamic-nameset",
     "DOM-class",
     "DOM-interface"]
     .map(function (cat) "JavaScript-" + cat) // Prevent spurious validator warnings
     .map(function (cat) [cat, cat.replace(/-/g, " ")]));

function iterCategory(category) {
    let enumerator = Services.catman.enumerateCategory(category);
    while (enumerator.hasMoreElements())
        yield enumerator.getNext().QueryInterface(Ci.nsISupportsCString).data;
}


function log(message, items, frame) {
    if (frame)
        Services.console.logMessage(
            Services.ScriptError(message + (items ? ": " + items.join(", ") : ""),
                                 frame.filename,
                                 null,
                                 frame.lineNumber,
                                 null,
                                 Ci.nsIScriptError.warningFlag,
                                 "chrome javascript exttest Extension Test"));
    else
        Services.console.logStringMessage(message + (items ? ": " + items.join(", ") : ""));

    if (frame && frame.lineNumber)
        message = frame.filename + ":" + frame.lineNumber + ": " + message
    dump(message + (items ? ":\n\t" + items.join("\n\t") : "") + "\n")
}

let xpath = function xpath(path, context) ({
    __proto__: (context.ownerDocument || context).evaluate(path, context, null, Ci.nsIDOMXPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null),
    __iterator__: function () { for (let i = 0; i < this.snapshotLength; i++) yield this.snapshotItem(i); }
});

let getOwnPropertyNames = Object.getOwnPropertyNames ||
    function debuggerProperties(obj) {
        if (Services.debugger.isOn) {
            let ret = {};
            Services.debugger.wrapValue(obj).getProperties(ret, {});
            return ret.value.map(function (prop) prop.name.stringValue);
        }
        return [];
    };

try {
    if (!Object.getOwnPropertyNames)
        Services.debugger.on();
}
catch (e) {}

